import numpy as np
import requests

from tclogger import StrsType, logger
from typing import Literal, TypedDict, Optional, Union

EmbedApiFormat = Literal["openai", "tei"]
EmbedResFormat = Literal["ndarray", "list2d"]


class EmbedClientConfigsType(TypedDict):
    endpoint: str
    api_key: Optional[str]
    model: str
    api_format: EmbedApiFormat = "tei"
    res_format: EmbedResFormat = "list2d"
    batch_size: int = 100


class EmbedClient:
    def __init__(
        self,
        endpoint: str,
        model: str = None,
        api_key: str = None,
        api_format: EmbedApiFormat = "tei",
        res_format: EmbedResFormat = "list2d",
        batch_size: int = 100,
        verbose: bool = False,
    ):
        self.endpoint = endpoint
        self.model = model
        self.api_key = api_key
        self.api_format = api_format
        self.res_format = res_format
        self.batch_size = batch_size
        self.verbose = verbose

    def log_resp_status(self, resp: requests.Response):
        err_mesg = f"× Embed error: {resp.status_code} {resp.text}"
        if self.verbose:
            logger.warn(err_mesg)
        raise ValueError(err_mesg)

    def log_embed_res(self, embeddings: list[list[float]]):
        if self.verbose:
            num = len(embeddings)
            dim = len(embeddings[0]) if num > 0 else 0
            val_type = type(embeddings[0][0]).__name__ if dim > 0 else "N/A"
            logger.okay(f"✓ Embed success: num={num}, dim={dim}, type={val_type}")

    def embed_batch(self, inputs: StrsType) -> list[list[float]]:
        headers = {
            "content-type": "application/json",
        }
        # https://huggingface.github.io/text-embeddings-inference/#/Text%20Embeddings%20Inference/embed
        payload = {
            "inputs": inputs,
            "normalize": True,
            "truncate": True,
        }
        resp = requests.post(self.endpoint, headers=headers, json=payload)
        if resp.status_code != 200:
            self.log_resp_status(resp)
            return []
        embeddings = resp.json()
        self.log_embed_res(embeddings)
        return embeddings

    def embed(self, inputs: StrsType) -> Union[list[list[float]], np.ndarray]:
        if isinstance(inputs, str):
            inputs = [inputs]
        if len(inputs) <= self.batch_size:
            return self.embed_batch(inputs)
        embs = []
        for i in range(0, len(inputs), self.batch_size):
            inputs_batch = inputs[i : i + self.batch_size]
            embs_batch = self.embed_batch(inputs_batch)
            embs.extend(embs_batch)
        if self.res_format == "ndarray":
            return np.array(embs)
        else:
            return embs


class EmbedClientByConfig(EmbedClient):
    def __init__(self, configs: EmbedClientConfigsType):
        super().__init__(**configs)



from .llm import LLMConfigsType, LLMClient, LLMClientByConfig
from .vector_utils import floats_to_bits, bits_to_hash, bits_dist, hash_dist
from .vector_utils import bits_sim, hash_sim, dot_sim
from .embed_client import EmbedClientConfigsType
from .embed_client import EmbedClient, EmbedClientByConfig
from .embed_server import TEIEmbedServerConfigsType
from .embed_server import TEIEmbedServer, TEIEmbedServerByConfig
from .embed_server import TEIEmbedServerArgParser, EmbedServerArgParser
from .gpu_fan import NvidiaSettingsParser, GPUFanController, GPUFanArgParser
from .gpu_fan import control_gpu_fan



import argparse
import shlex

from pathlib import Path
from tclogger import logger, shell_cmd
from typing import TypedDict


class TEIEmbedServerConfigsType(TypedDict):
    port: int
    model_name: str
    instance_id: str
    verbose: bool


class TEIEmbedServer:
    def __init__(
        self,
        port: int = None,
        model_name: str = None,
        instance_id: str = None,
        hf_token: str = None,
        verbose: bool = False,
    ):
        self.port = port
        self.model_name = model_name
        self.instance_id = instance_id or self.default_instance_id()
        self.hf_token = hf_token
        self.verbose = verbose

    def default_instance_id(self) -> str:
        return self.model_name.replace("/", "--")

    def run(self):
        script_path = Path(__file__).resolve().parent / "run_tei.sh"
        if not script_path.exists():
            logger.warn(f"× Missing `run_tei.sh`: {script_path}")
            return

        run_parts = ["bash", str(script_path)]
        if self.port:
            run_parts.extend(["-p", str(self.port)])
        if self.model_name:
            run_parts.extend(["-m", self.model_name])
        if self.instance_id:
            run_parts.extend(["-id", self.instance_id])
        if self.hf_token:
            run_parts.extend(["-u", self.hf_token])
        cmd_run = shlex.join(run_parts)
        shell_cmd(cmd_run)

        if self.verbose:
            cmd_logs = f'docker logs -f --tail 0 "{self.instance_id}"'
            shell_cmd(cmd_logs)

    def kill(self):
        if not self.instance_id:
            logger.warn("× Missing arg: -id (--instance-id)")
            return

        cmd_kill = f'docker stop "{self.instance_id}"'
        shell_cmd(cmd_kill)

    def remove(self):
        if not self.instance_id:
            logger.warn("× Missing arg: -id (--instance-id)")
            return

        cmd_rm = f'docker rm -f "{self.instance_id}"'
        shell_cmd(cmd_rm)


class TEIEmbedServerByConfig(TEIEmbedServer):
    def __init__(self, configs: TEIEmbedServerConfigsType):
        super().__init__(**configs)


class TEIEmbedServerArgParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_argument("-m", "--model-name", type=str, default=None)
        self.add_argument("-id", "--instance-id", type=str, default=None)
        self.add_argument("-u", "--hf-token", type=str, default=None)
        self.add_argument("-p", "--port", type=int, default=28888)
        self.add_argument("-b", "--verbose", action="store_true")
        self.add_argument("-rm", "--remove", action="store_true")
        self.add_argument("-k", "--kill", action="store_true")
        self.args, _ = self.parse_known_args()


class EmbedServerArgParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_argument("-t", "--type", type=str, choices=["tei"], default="tei")
        self.args, _ = self.parse_known_args()


def main():
    main_args = EmbedServerArgParser().args
    if main_args.type == "tei":
        args = TEIEmbedServerArgParser().args
        embed_server = TEIEmbedServer(
            port=args.port,
            model_name=args.model_name,
            instance_id=args.instance_id,
            hf_token=args.hf_token,
            verbose=args.verbose,
        )
        if args.remove:
            embed_server.remove()
        elif args.kill:
            embed_server.kill()
        else:
            embed_server.run()


if __name__ == "__main__":
    main()

    # Case 1: gte-multilingual-base
    # python -m tfmx.embed_server -t "tei" -m "Alibaba-NLP/gte-multilingual-base" -p 28888 -b
    # python -m tfmx.embed_server -t "tei" -m "Alibaba-NLP/gte-multilingual-base" -k

    # Case 2: bge-large-zh-v1.5
    # python -m tfmx.embed_server -t "tei" -m "BAAI/bge-large-zh-v1.5" -p 28889 -b
    # python -m tfmx.embed_server -t "tei" -m "BAAI/bge-large-zh-v1.5" -k

    # Case 3: Qwen/Qwen3-Embedding-0.6B
    # python -m tfmx.embed_server -t "tei" -m "Qwen/Qwen3-Embedding-0.6B" -p 28887 -b
    # python -m tfmx.embed_server -t "tei" -m "Qwen/Qwen3-Embedding-0.6B" -k
    # python -m tfmx.embed_server -t "tei" -m "Qwen/Qwen3-Embedding-0.6B" -rm

    # Case 4: Use HF_TOKEN
    # python -m tfmx.embed_server -t "tei" -m "Qwen/Qwen3-Embedding-0.6B" -p 28887 -b -u hf_****Y



import asyncio
import ast
import json
import re
import requests

from tclogger import logger, logstr, Runtimer
from tclogger import dict_to_str, dt_to_str, obj_params
from typing import Literal, TypedDict

LlmApiType = Literal["openai", "ollama", "doubao"]


class LLMConfigsType(TypedDict):
    endpoint: str
    api_key: str
    model: str
    api_format: LlmApiType = "openai"
    stream: bool = None
    init_messages: list = []
    enable_thinking: bool = None
    delta_func: callable = None
    terminate_event: asyncio.Event = None
    verbose_user: bool = True
    verbose_assistant: bool = True
    verbose_think: bool = True
    verbose_content: bool = True
    verbose_usage: bool = True
    verbose_finish: bool = True
    verbose: bool = False


DEFAULT_CHAT_PARAMS = {
    "model": "",
    "stream": True,
    "temperature": 0.0,
}


class LLMClient:
    def __init__(
        self,
        endpoint: str,
        api_key: str,
        api_format: LlmApiType = "openai",
        model: str = None,
        stream: bool = None,
        init_messages: list = [],
        enable_thinking: bool = None,  # used by qwen3 and doubao
        delta_func: callable = None,
        terminate_event: asyncio.Event = None,
        verbose_user: bool = True,
        verbose_assistant: bool = True,
        verbose_content: bool = True,
        verbose_think: bool = True,
        verbose_usage: bool = True,
        verbose_finish: bool = True,
        verbose: bool = False,
    ):
        self.endpoint = endpoint
        self.api_key = api_key
        self.api_format = api_format
        self.model = model
        self.stream = stream
        self.init_messages = init_messages
        self.enable_thinking = enable_thinking
        self.delta_func = delta_func
        self.terminate_event = terminate_event

        self.verbose_user = verbose_user
        self.verbose_assistant = verbose_assistant
        self.verbose_content = verbose_content
        self.verbose_think = verbose_think
        self.verbose_usage = verbose_usage
        self.verbose_finish = verbose_finish
        self.verbose = verbose

        self.is_thinking = False

    def set_enable_thinking(self, enable_thinking: bool = None) -> bool:
        if enable_thinking is not None:
            self.enable_thinking = enable_thinking
        return self.enable_thinking

    def set_think_status(self, tag: str = "<think>"):
        if self.enable_thinking is not False and not self.is_thinking:
            self.is_thinking = True
            logger.mesg(tag, end="", verbose=self.verbose_think)

    def reset_think_status(self, tag: str = "</think>"):
        if self.enable_thinking is not False and self.is_thinking:
            self.is_thinking = False
            logger.mesg(tag, end="", verbose=self.verbose_think)

    def create_response(
        self,
        messages: list,
        model: str = None,
        enable_thinking: bool = None,
        temperature: float = None,
        seed: int = None,
        stream: bool = None,
    ):
        headers = {
            "accept": "application/json",
            "content-type": "application/json",
        }
        if self.api_key:
            headers["authorization"] = f"Bearer {self.api_key}"
        model, stream = obj_params(
            self, DEFAULT_CHAT_PARAMS, model=model, stream=stream
        )
        enable_thinking = self.set_enable_thinking(enable_thinking)
        payload = {
            "model": model,
            "messages": self.init_messages + messages,
            "stream": stream,
        }
        options = {}
        if temperature is not None:
            options["temperature"] = temperature
        if seed is not None:
            options["seed"] = seed
        if self.api_format == "ollama":
            payload["options"] = options
        else:
            payload.update(options)
        if enable_thinking is not None:
            if self.api_format == "doubao":
                # https://www.volcengine.com/docs/82379/1585128
                if enable_thinking:
                    thinking_type = "enabled"
                else:
                    thinking_type = "disabled"
                payload["thinking"] = {"type": thinking_type}
            else:
                payload["chat_template_kwargs"] = {"enable_thinking": enable_thinking}
        response = requests.post(
            self.endpoint, headers=headers, json=payload, stream=stream
        )
        return response

    def exec_delta_func(self, role: str, content: str):
        if self.delta_func:
            delta_func_args = {"role": role, "content": content}
            if asyncio.iscoroutinefunction(self.delta_func):
                asyncio.run(self.delta_func(**delta_func_args))
            else:
                self.delta_func(**delta_func_args)

    def parse_stream_response(self, response: requests.Response) -> tuple[str, dict]:
        response_content = ""
        usage = None
        for line in response.iter_lines():
            if self.terminate_event and self.terminate_event.is_set():
                break

            line = line.decode("utf-8")
            remove_patterns = [r"^\s*data:\s*", r"^\s*\[DONE\]\s*"]
            for pattern in remove_patterns:
                line = re.sub(pattern, "", line).strip()

            if line:
                try:
                    line_data = json.loads(line)
                except Exception as e:
                    try:
                        line_data = ast.literal_eval(line)
                    except:
                        logger.warn(f"× Error: {line}")
                        logger.err(e)
                        raise e

                if self.api_format == "ollama":
                    # https://github.com/ollama/ollama/blob/main/docs/api.md#response-9
                    delta_data = line_data["message"]
                    finish_reason = "stop" if line_data["done"] else None
                else:
                    # https://platform.openai.com/docs/api-reference/chat/streaming
                    delta_data = line_data["choices"][0]["delta"]
                    finish_reason = line_data["choices"][0].get("finish_reason", None)
                if "role" in delta_data:
                    role = delta_data["role"]
                delta_reasoning_content = delta_data.get("reasoning_content", None)
                if delta_reasoning_content is not None:
                    self.set_think_status()
                    delta_content = delta_reasoning_content
                    if delta_content:
                        response_content += delta_content
                    logger.mesg(delta_content, end="", verbose=self.verbose_content)
                delta_content = delta_data.get("content", None)
                if delta_content is not None:
                    if delta_content or delta_reasoning_content is None:
                        self.reset_think_status()
                    try:
                        if delta_content:
                            response_content += delta_content
                    except Exception as e:
                        logger.warn(delta_data)
                        logger.err(e)
                        raise e
                    logger.mesg(delta_content, end="", verbose=self.verbose_content)
                    self.exec_delta_func(role, delta_content)
                if finish_reason == "stop":
                    if line_data.get("usage", {}):
                        usage = line_data["usage"]
                        logger.file(
                            "\n" + dict_to_str(usage), verbose=self.verbose_usage
                        )
                    logger.success("\n[Finished]", end="", verbose=self.verbose_finish)
                    self.exec_delta_func("stop", "")

        return response_content, usage

    def parse_json_response(self, response: requests.Response) -> tuple[str, dict]:
        response_content = ""
        usage = None
        try:
            response_data = response.json()
            response_content = ""
            if self.api_format == "ollama":
                message = response_data["message"]
            else:
                message = response_data["choices"][0]["message"]
            reasoning_content = message.get("reasoning_content", "")
            content = message.get("content", "")
            if reasoning_content and content:
                response_content = f"<think>{reasoning_content}</think>" + content
            elif reasoning_content:
                response_content = reasoning_content
            else:
                response_content = content
            if "usage" in response_data:
                usage = response_data["usage"]
                if usage and self.verbose_usage:
                    logger.file("\n" + dict_to_str(usage))
            if self.verbose_content:
                logger.mesg(response_content)
            if self.verbose_finish:
                logger.success("[Finished]", end="")
        except Exception as e:
            logger.warn(f"× Error: {response.text}")
        return response_content, usage

    def chat(
        self,
        messages: list,
        model: str = None,
        enable_thinking: bool = None,
        temperature: float = None,
        seed: int = None,
        stream: bool = None,
        verbose: bool = None,
    ) -> str:
        timer = Runtimer(verbose=False)
        timer.start_time()
        model, stream = obj_params(
            self, DEFAULT_CHAT_PARAMS, model=model, stream=stream
        )

        verbose = verbose if verbose is not None else self.verbose
        logger.enter_quiet(not verbose)

        if self.verbose_user:
            try:
                user_prompt = messages[-1]["content"]
                logger.note(f"USER: {user_prompt}")
            except Exception as e:
                logger.warn(messages)
                logger.err(e)
                raise e

        response = self.create_response(
            messages=messages,
            model=model,
            enable_thinking=enable_thinking,
            temperature=temperature,
            seed=seed,
            stream=stream,
        )

        if self.verbose_assistant:
            logger.mesg("ASSISTANT: ", end="")

        if stream:
            response_content, usage = self.parse_stream_response(response)
        else:
            response_content, usage = self.parse_json_response(response)
        timer.end_time()
        if self.verbose_finish:
            elapsed_time = dt_to_str(
                timer.elapsed_time(), precision=1, str_format="unit"
            )
            model_name_str = "[" + model.split("/")[-1] + "]"
            logger.note(f" ({elapsed_time}) {logstr.file(model_name_str)}")
        else:
            logger.note("", verbose=self.verbose_content)

        logger.exit_quiet(not verbose)
        return response_content


class LLMClientByConfig(LLMClient):
    def __init__(self, configs: LLMConfigsType):
        super().__init__(**configs)
